/*
 * Copyright (C) 2001 Stealth.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Stealth.
 * 4. The name Stealth may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <vector>
extern "C" {
#include <openssl/md5.h>
}
#include "chap.h"

void die(const char *s)
{
	perror(s);
	exit(errno);
}


void help()
{
	printf("Usage: crc [-d] [-M] <-f file>\n\n"
		"I expect passwords on stdin.\n\n");
	exit(1);
}


void dump(const char *file)
{
	int fd;

	if ((fd = open(file, O_RDONLY)) < 0)
		die("open");
	
	chap_session s;
	while (read(fd, &s, sizeof(s)) == sizeof(s)) {
		printf("******* %s\nChallenge: 0x", s.login);
		for (u_int i = 0; i < s.chal_len; ++i)
			printf("%02x", s.challenge[i]);
		printf("\nResponse: 0x");
		for (u_int i = 0; i < s.resp_len; ++i)
			printf("%02x", s.response[i]);
		printf("\nType: ");
		if (s.chal_type == CHAP_DIGEST_MD5)
			printf("MD5");
		else if (s.chal_type == CHAP_MICROSOFT) 
			printf("MS-CHAP");
		else if (s.chal_type == CHAP_MICROSOFT_V2)
			printf("MS-CHAPv2");
		else
			printf("???\n");
		printf("\n\n");
	}
}


void crackMD5(const char *file)
{
	int fd;
	vector<chap_session> v;

	if ((fd = open(file, O_RDONLY)) < 0)
		die("open");
	
	chap_session s;
	u_short howmuch = 0;
	while (read(fd, &s, sizeof(s)) == sizeof(s)) {
		if (s.chal_type == CHAP_DIGEST_MD5) {
			v.push_back(s);
			++howmuch;
		}
	}
	close(fd);

	printf("Found %d matching records.\n", howmuch);

	MD5_CTX md;
	u_char digest[MD5_SIGNATURE_SIZE];
	char word[15];
	vector<chap_session>::iterator i;
	while (fgets(word, sizeof(word), stdin) != NULL) {
		word[strlen(word)-1] = 0;
		for (i = v.begin(); i != v.end(); ++i) {
			MD5_Init(&md);
			MD5_Update(&md, &i->id, 1);
			MD5_Update(&md, (unsigned char*)word, strlen(word));
			MD5_Update(&md, i->challenge, i->chal_len);
			MD5_Final(digest, &md);
			if (memcmp(digest, i->response, MD5_SIGNATURE_SIZE) == 0)
				printf("\a-> %s/%s <-\n", i->login, word);
		}
		memset(word, 0, sizeof(word));
	}
}

int main(int argc, char **argv)
{
	int c, _dump = 0, _crackMD5 = 0;
	char *file = NULL;

	while ((c = getopt(argc, argv, "dMw:f:")) != -1) {
		switch (c) {
		case 'd':
			++_dump;
			break;
		case 'M':
			++_crackMD5;
			break;
		case 'f':
			file = strdup(optarg);
			break;
		default:
			break;
		}
	}

	if (_crackMD5 && file)
		crackMD5(file);
	else if (_dump && file)
		dump(file);
	else
		help();
	return 0;
}

